import asyncio
import pyppeteer

class Conversation:
    unread: bool 
    id: int
    timestamp: str 
    fromNumber: str
    latestMsgText: str

    def __init__(self, unread, id, timestamp, fromNumber, latestMsgText) -> None:
        self.unread = unread
        self.id = id
        self.timestamp = timestamp
        self.fromNumber = fromNumber
        self.latestMsgText = latestMsgText

    

class MessageService: 
    page = pyppeteer.page.Page

    def __init__(self, page: pyppeteer.page.Page):
        self.page = page

    async def getInbox(self):
        # TODO: add pagination
        await self.page.waitForNavigation({"waitUntil": 'load'})
        await self.page.waitForSelector('body > mw-app > mw-bootstrap > div > main > mw-main-container > div > mw-main-nav > mws-conversations-list > nav > div.conv-container.ng-star-inserted > mws-conversation-list-item')

        inbox = await self.page.evaluate("""() => {
        
            function evalConvoElement (conversation) {
                const props = {
                    unread: false, // querySelector find .unread class
                    id: 0, // href of a tag
                    timestamp: '', // mws-relative-timestamp .innerText || > ..ng-star-inserted').getAttribute('aria-label') if latest message
                    from: '', // querySelector('h3').innerText
                    latestMsgText: '' // querySelector('mws-conversation-snippet').innerText
                }
                props.unread = conversation.querySelector('.unread') ? true : false
                
                const regex = /conversations\/(\d{1,})/g
                const chatUrl = conversation.querySelector('a').href
                props.id = parseInt(chatUrl.match(regex)[0].split('conversations/')[1])
                
                if (conversation.querySelector('mws-relative-timestamp').childElementCount > 0) {
                    props.timestamp = conversation.querySelector('mws-relative-timestamp > .ng-star-inserted').getAttribute('aria-label')
                } else {
                    props.timestamp = (conversation.querySelector('mws-relative-timestamp')).innerText
                }

                props.from = conversation.querySelector('h3').innerText
                props.latestMsgText = (conversation.querySelector('mws-conversation-snippet')).innerText
                if (props.latestMsgText.startsWith('You:')) {
                    props.latestMsgText = props.latestMsgText.slice('You:'.length).trim()
                }
                return props
            }

            const conversations = document.querySelectorAll("body > mw-app > mw-bootstrap > div > main > mw-main-container > div > mw-main-nav > mws-conversations-list > nav > div.conv-container.ng-star-inserted > mws-conversation-list-item")
            const msgs = []
            for (const conversation of conversations) {
                if (conversation) {
                    msgs.push(evalConvoElement(conversation))
                }
            }
            return msgs
        }
        """)
        return inbox

    async def sendMessage(self,to:str, text:str):
        pass





import asyncio
import json
import signal
from typing import Callable

import pyppeteer
from pyee import AsyncIOEventEmitter

from pymessages.service import MessageService


class ClientOptions: 
    headless:bool
    credentials: object

    def __init__(self,headless=False, credentials={"cookies":[], "localStorage": {}}):
        self.headless = headless
        self.credentials = credentials

class Listener:
    group: int
    func: Callable
    event: str

    def __init__(self, group, func, event) -> None:  #TODO add exceptions
        self.group = group
        self.func = func
        self.event = event

    def check(self, event) -> bool:
        return self.event == event

    def call(self, data, service) -> object:  #TODO add exceptions
        return self.func(data, service)

    def callonevent(self, event, data, service) -> object: 
        if self.event == event:
            self.func(data, service)
        return None

class MessagesClient(AsyncIOEventEmitter):
    page: pyppeteer.page.Page
    browser: pyppeteer.browser.Browser
    groups = []
    listeners = {}
    options: ClientOptions
    is_authenticated: bool = False

    def __init__(self, credentials={"cookies":[], "localStorage": {}}, headless=True):
        super(MessagesClient, self).__init__()
        self.loop = asyncio.get_event_loop()
        try:
            self.loop.add_signal_handler(signal.SIGTERM, self.stop)
        except NotImplementedError:
            pass
        self.options = ClientOptions(headless, credentials)

    def stop(self):
        print('Pymessages stopping.')
        self.loop.stop()
        
    def launch(self):
        self.loop.run_until_complete(self._launch(self.options))

    def idle(self, close=True):
        self.loop.run_forever()


    @staticmethod
    def loadCredentialFile(path):
        with open(path) as f:
            cred = json.load(f)
        return cred

    async def _attachReqTracer(self):
        @self.page.on('request')
        def on_request(request):
            url = request.url
            # print("Check ------")
            if "Pairing/GetWebEncryptionKey" in url:
                # print("YESSSSS")
                service = MessageService(self.page)
                if not self.is_authenticated:
                    self.emit('authenticated', service) #TODO: pass credentials as well
                    self.is_authenticated = True

    
    async def _attachQrReader(self):
        await self.page.waitForSelector("body > mw-app > mw-bootstrap > div > main > mw-authentication-container > div > div.content-container > div > div.qr-code-container > div.qr-code-wrapper > mw-qr-code")
        async def _func_to_expose():
            img = await self.page.J('body > mw-app > mw-bootstrap > div > main > mw-authentication-container > div > div.content-container > div > div.qr-code-container > div.qr-code-wrapper > mw-qr-code > img')
            if img:
                src = await img.getProperty('src')
                if src:
                    self.emit('qr-code', await src.jsonValue())  # qrData = base64 qr image
        await self.page.exposeFunction('onQrChange', _func_to_expose)
        await self.page.evaluate(""" () => {
            const observer = new MutationObserver((mutations) => {
                for (const mutation of mutations) {
                    if (mutation.attributeName === 'data-qr-code') {
                        // @ts-ignore
                        window.onQrChange(mutation)
                    }
                }
            })
            const img = document.querySelector("body > mw-app > mw-bootstrap > div > main > mw-authentication-container > div > div.content-container > div > div.qr-code-container > div.qr-code-wrapper > mw-qr-code")
            if (img) {
                observer.observe(img, { attributes: true, childList: true, characterData: true })
            }
            return observer
        }
        """)
        await self.page.waitForSelector('body > mw-app > mw-bootstrap > div > main > mw-authentication-container > div > div.content-container > div > div.qr-code-container > div.qr-code-wrapper > mw-qr-code > img')
        img = await self.page.J('body > mw-app > mw-bootstrap > div > main > mw-authentication-container > div > div.content-container > div > div.qr-code-container > div.qr-code-wrapper > mw-qr-code > img')
        if img:
            src = await img.getProperty("src")
            if src:
                self.emit('qr-code', await src.jsonValue())

    async def _launch(self, options: ClientOptions):
        browser = await pyppeteer.launch({"headless": options.headless})
        self.browser = browser
        page = await browser.newPage()
        self.page = page
        await self.page.goto('https://messages.android.com', { "waitUntil": 'load' })
        await self.page.waitForSelector('#mat-slide-toggle-1-input')
        await self.page.evaluate("""() => {
            const checkbox = document.querySelector('#mat-slide-toggle-1-input')
            checkbox.click() 
        }
        """) #remember me button
        self.emit('browser-launched')
        if len(options.credentials["localStorage"]) == 0:
            await self._attachQrReader()
            await self._attachReqTracer()
            return
        else:
            await self.setCredentials(options.credentials)
            service = MessageService(self.page)
            self.emit('authenticated', service)
            self.is_authenticated = True

    async def getCredentials(self):
        await self.page.waitForFunction('!!localStorage.getItem("pr_backend_type")')
        localStorageData = await self.page.evaluate("""() => {
            let data = {}
            Object.assign(data, window.localStorage)
            return data
        }
        """)
        cookiz = await self.page.cookies()
        creds = {
            "cookies": cookiz,
            "localStorage": localStorageData
        }
        return creds

    async def setCredentials(self, credentials: dict):
        await self.page.setCookie(*credentials["cookies"])
        await self.page.evaluate("""(localStorageData) => {  
            try {
                localStorageData = JSON.parse(localStorageData)
            } catch (err) {}
            for (const key of Object.keys(localStorageData)) {
                localStorage.setItem(key, localStorageData[key])
            }
        }""", json.dumps(credentials["localStorage"]))
        await self.page.reload()
        return


    async def quit(self):
        await self.browser.close()
        

    def __del__(self):
        self.loop.run_until_complete(self.quit())




from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='pymessages',
      version='0.1',
      description='This is a Google Messages Client library to send message with a backend .eg. with flask to send otp messages. This module uses your own number to work as a sms gateway api and you can send message to other person with your own number.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: ISC License (ISCL)',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries',
      ],
      keywords='python sms-api sms-gateway google-messages',
      url='http://github.com/shivamsn97/pymessages',
      author='Shivam Saini',
      author_email='shivamsn97@gmail.com',
      license='ISC',
      packages=find_packages(),
      install_requires=[
          'pyppeteer',
          'pyee',
      ],
      include_package_data=True,
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose'],
      )
